# Author: Robert Park <robru@gottengeography.ca>, (C) 2011
# Copyright: See COPYING file included with this distribution.
#
# Package version data.
#
# Please note that this file is equally valid as Python as it is Bash,
# as it is also imported by the po/initialize-pot.sh script.

APPNAME='GottenGeography'
PACKAGE='gottengeography'
VERSION='2.5'

AUTHOR='Robert Park'
EMAIL='robru@gottengeography.ca'



# Author: Robert Park <robru@gottengeography.ca>, (C) 2012
# Copyright: See COPYING file included with this distribution.

"""Orchestrate the construction of widgets with GtkBuilder.

>>> type(Widgets.main)
<class 'gi.repository.Gtk.ApplicationWindow'>
>>> type(Widgets['error_bar'])
<class 'gi.repository.Gtk.InfoBar'>

>>> Widgets.status_message('Hello!', True)
>>> Widgets.error_message.get_text()
'Hello!'
"""


from gi.repository import GtkChamplain, Champlain
from gi.repository import Gdk, GdkPixbuf
from gi.repository import Gtk, GLib
from time import gmtime, strftime
from os.path import join

from gg.version import APPNAME, PACKAGE
from gg.build_info import PKG_DATA_DIR, REVISION
from gg.common import Gst, singleton, memoize
from gg.common import modified, selected


CONTROL_MASK = Gdk.ModifierType.CONTROL_MASK
SHIFT_MASK = Gdk.ModifierType.SHIFT_MASK


class Builder(Gtk.Builder):
    """Load GottenGeography's UI definitions."""

    def __init__(self, filename=PACKAGE):
        Gtk.Builder.__init__(self)

        self.set_translation_domain(PACKAGE)
        self.add_from_file(join(PKG_DATA_DIR, filename + '.ui'))

    @memoize
    def __getattr__(self, widget):
        """Make calls to Gtk.Builder().get_object() more pythonic.

        Here is a quick comparison of execution performance:

        Executing this method, no memoize:  6.50 microseconds
        Calling get_object directly:        4.35 microseconds
        Executing this method with memoize: 1.34 microseconds
        Accessing an instance attribute:    0.08 microseconds
        Accessing a local variable:         0.03 microseconds

        (averaged over a million executions with the timeit package)

        Considering that this method is 3 orders of magnitude slower than
        accessing variables, you should avoid it inside performance-critical
        inner loops, however thanks to memoization, it's faster than calling
        get_object() directly, so don't sweat it.

        >>> infobar = 'error_bar'
        >>> Widgets.error_bar is Widgets.get_object(infobar)
        True
        >>> Widgets.error_bar is Widgets[infobar]
        True
        >>> Widgets.error_bar is Builder().error_bar
        False
        """
        return self.get_object(widget)

    __getitem__ = __getattr__


@singleton
class Widgets(Builder):
    """Tweak the GtkBuilder results specifically for the main window."""
    message_timeout_source = None
    defer_select = False

    def __init__(self):
        Builder.__init__(self)

    def launch(self):
        """Do some things that GtkBuilder XML can't do.

        Ideally this method would not exist. If you see something here that
        can be done directly in the GtkBuilder XML, please let me know.
        """
        self.loaded_photos.set_sort_column_id(3, Gtk.SortType.ASCENDING)

        self.about.set_version(str(REVISION))
        self.about.set_program_name(APPNAME)
        self.about.set_logo(GdkPixbuf.Pixbuf.new_from_file_at_size(
            join(PKG_DATA_DIR, PACKAGE + '.svg'), 192, 192))

        self.main.resize(*Gst.get('window-size'))
        self.main.show_all()

        Gst.bind('left-pane-page', self.photo_camera_gps, 'page')
        Gst.bind('use-dark-theme', Gtk.Settings.get_default(),
                 'gtk-application-prefer-dark-theme')

        placeholder = self.empty_photo_list
        toolbar = self.photo_btn_bar

        def photo_pane_visibility(liststore, *ignore):
            """Hide the placeholder and show the toolbar when appropriate."""
            empty = liststore.get_iter_first() is None
            placeholder.set_visible(empty)
            toolbar.set_visible(not empty)

        self.loaded_photos.connect('row-inserted', photo_pane_visibility)
        self.loaded_photos.connect('row-deleted', photo_pane_visibility)
        self.photos_selection.connect('changed', self.update_highlights)
        self.photos_selection.connect('changed', self.button_sensitivity)
        self.photos_view.connect('button-press-event', self.photoview_pressed)
        self.photos_view.connect(
            'button-release-event', self.photoview_released)
        self.photos_view.connect('row-activated', self.show_large_preview)
        self.large_preview_window.connect(
            'delete_event', lambda *ignore:
                self.large_preview_window.hide_on_delete())

        self.error_bar.connect(
            'response', lambda widget, signal: widget.hide())

    def show_large_preview(self, view, path, column, handler_ids=set()):
        """Show the large preview window."""
        photo = selected.copy().pop()
        stamp = gmtime(photo.timestamp)  # UTC

        self.clock_photo_hours.set_value(stamp.tm_hour)
        self.clock_photo_minutes.set_value(stamp.tm_min)
        self.clock_photo_seconds.set_value(stamp.tm_sec)
        self.clock_photo_tz.set_active_id('0.0')
        self.large_preview.set_from_pixbuf(photo.get_large_preview())

        while handler_ids:
            self.clock_photo_button.disconnect(handler_ids.pop())
        handler_ids.add(self.clock_photo_button.connect(
            'clicked', photo.camera.get_offset_from_clock_photo,
            stamp, strftime('%z')))

        self.large_preview_window.show_all()
        self.large_preview_window.present()

    def update_highlights(self, selection):
        """Ensure only the selected labels are highlighted."""
        selection_exists = selection.count_selected_rows() > 0
        selected.clear()
        for label in MarkerLayer.get_markers():
            photo = label.photo
            itr = photo.iter
            if itr and selection.iter_is_selected(itr):
                selected.add(photo)
            label.set_highlight(photo in selected, selection_exists)

    def button_sensitivity(self, *ignore):
        """Control the sensitivity of various widgets."""
        self.close_button.set_sensitive(selected)
        self.apply_button.set_sensitive(selected)
        self.save_button.set_sensitive(modified)
        self.revert_button.set_sensitive(modified & selected)
        self.jump_button.set_sensitive(
            [photo for photo in selected if photo.positioned])

    # Multiple selection drag and drop copied from Kevin Mehall, adapted
    # to use it with the standard GtkTreeView.
    # http://blog.kevinmehall.net/2010/pygtk_multi_select_drag_drop
    def photoview_pressed(self, tree, event):
        """Allow drag & drop with multiple photos selected."""
        target = tree.get_path_at_pos(int(event.x), int(event.y))
        if (target and event.type == Gdk.EventType.BUTTON_PRESS and
                not (event.state & (CONTROL_MASK | SHIFT_MASK)) and
                self.photos_selection.path_is_selected(target[0])):
            # disable selection
            self.photos_selection.set_select_function(
                lambda *ignore: False, None)
            self.defer_select = target[0]

    def photoview_released(self, tree, event):
        """Restore normal selection behavior while not dragging."""
        self.photos_selection.set_select_function(lambda *ignore: True, None)
        target = tree.get_path_at_pos(int(event.x), int(event.y))
        if (target and self.defer_select and
                self.defer_select == target[0] and
                not (event.x == 0 and event.y == 0)):  # drag & drop
            tree.set_cursor(target[0], target[1], False)

    def redraw_interface(self, fraction=None, text=None):
        """Tell Gtk to redraw the user interface, so it doesn't look hung.

        Primarily for updating the progressbar, but won't touch it if called
        with no arguments.
        """
        if fraction is not None:
            self.progressbar.set_fraction(fraction)
        if text is not None:
            self.progressbar.set_text(str(text))
        while Gtk.events_pending():
            Gtk.main_iteration()

    def dismiss_message(self):
        """Responsible for hiding the GtkInfoBar after a timeout."""
        self.message_timeout_source = None
        self.error_bar.hide()
        return False

    def status_message(self, message, info=False):
        """Display a message with the GtkInfoBar."""
        self.error_message.set_markup('<b>{}</b>'.format(message))
        self.error_bar.set_message_type(
            Gtk.MessageType.INFO if info else Gtk.MessageType.WARNING)
        self.error_icon.set_from_stock(
            Gtk.STOCK_DIALOG_INFO if info else Gtk.STOCK_DIALOG_WARNING, 6)
        self.error_bar.show()

        # Remove any previous message timeout
        if self.message_timeout_source is not None:
            GLib.source_remove(self.message_timeout_source)
        if info:
            self.message_timeout_source = \
                GLib.timeout_add_seconds(10, self.dismiss_message)


@singleton
class ChamplainEmbedder(GtkChamplain.Embed):
    """Put the map view onto the main window."""

    def __init__(self):
        GtkChamplain.Embed.__init__(self)
        Widgets.map_container.add(self)


# Just pretend that MapView is also a singleton...
MapView = ChamplainEmbedder.get_view()


@singleton
class MarkerLayer(Champlain.MarkerLayer):
    """This is the primary MarkerLayer upon which all Photo labels rest."""

    def __init__(self):
        Champlain.MarkerLayer.__init__(self)
        MapView.add_layer(self)



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Main application code that ties all the other modules together."""


from gg.version import APPNAME, PACKAGE

import gettext
gettext.bindtextdomain(PACKAGE)
gettext.textdomain(PACKAGE)

from gi.repository import GLib, GObject, GtkClutter, Gtk, Gdk, Gio
from os.path import basename, abspath
from gettext import gettext as _

# If I have seen a little further it is by standing on the shoulders of Giants.
#                                    --- Isaac Newton

if not GLib.get_application_name():
    GLib.set_application_name(APPNAME)
    GObject.set_prgname(PACKAGE)

GtkClutter.init([])

from gg.camera import Camera
from gg.xmlfiles import TrackFile
from gg.gpsmath import Coordinates
from gg.widgets import Widgets, MapView
from gg.actor import CoordLabel, animate_in
from gg.photos import Photograph, fetch_thumbnail
from gg.navigation import go_back, move_by_arrow_keys
from gg.common import Gst, Binding, selected, modified

from gg.drag import DragController
from gg.search import SearchController

# Handy names for GtkListStore column numbers.
PATH, SUMMARY, THUMB, TIMESTAMP = range(4)

# Just pretend these functions are actually GottenGeography() instance methods.
# The 'self' argument gets passed in by GtkApplication instead of Python.


def command_line(self, commands):
    """Open the files passed in at the commandline.

    This method collects any commandline arguments from any invocation of
    GottenGeography and reports them to the primary instance for opening.
    """
    files = commands.get_arguments()[1:]
    if files:
        self.activate()
        self.open_files([abspath(f) for f in files])
    return 0


def startup(self):
    """Display the primary window and connect some signals."""
    self.quit_message = Widgets.quit.get_property('secondary-text')

    self.drag = DragController(self.open_files)
    self.search = SearchController()

    center = Coordinates()
    Binding(MapView, 'latitude', center)
    Binding(MapView, 'longitude', center)
    center.do_modified()
    Binding(center, 'geoname', Widgets.main, 'title')
    Binding(center, 'coords', CoordLabel, 'text')
    center.timeout_seconds = 10  # Only update titlebar every 10 seconds

    # Toolbar bindings.
    click_handlers = {
        'open':
            self.add_files_dialog,
        'save':
            self.save_all_files,
        'close':
            lambda btn: [p.destroy() for p in selected.copy()],
        'revert':
            lambda btn: self.open_files(
                [p.filename for p in modified & selected]),
        'about':
            lambda *ignore: Widgets.about.run() and Widgets.about.hide(),
        'help':
            self.open_help,
        'jump':
            self.jump_to_photo,
        'apply':
            self.apply_selected_photos,
        'map_source_menu':
            self.map_source_menu,
    }
    for name, handler in click_handlers.items():
        button = Widgets[name + '_button']
        if button:
            button.connect('clicked', handler)

    Widgets.zoom_in_button.connect('clicked', lambda *x: MapView.zoom_in())
    Widgets.zoom_out_button.connect('clicked', lambda *x: MapView.zoom_out())
    Widgets.back_button.connect('clicked', go_back)

    Widgets.open.connect(
        'update-preview', self.update_preview, Widgets.preview)

    accel = Gtk.AccelGroup()
    accel.connect(Gdk.keyval_from_name('q'),
                  Gdk.ModifierType.CONTROL_MASK, 0, self.confirm_quit_dialog)

    for key in ('Left', 'Right', 'Up', 'Down'):
        accel.connect(
            Gdk.keyval_from_name(key),
            Gdk.ModifierType.MOD1_MASK,
            0,
            move_by_arrow_keys)

    Widgets.main.add_accel_group(accel)
    Widgets.main.connect('delete_event', self.confirm_quit_dialog)
    self.add_window(Widgets.main)

    save_size = lambda v, s, size: Gst.set_window_size(size())
    for prop in ['width', 'height']:
        MapView.connect('notify::' + prop, save_size, Widgets.main.get_size)

    Widgets.button_sensitivity()

    Gst.connect('changed::thumbnail-size', Photograph.resize_all_photos)

    Widgets.launch()
    animate_in(self.do_fade_in)


class GottenGeography(Gtk.Application):
    """Provides a graphical interface to automagically geotag photos.

    Just load your photos, and load a GPX file, and GottenGeography will
    automatically cross-reference the timestamps on the photos to the
    timestamps in the GPX to determine the three-dimensional coordinates of
    each photo.
    """

    def __init__(self, do_fade_in=True):
        Gtk.Application.__init__(
            self, application_id='ca.' + APPNAME,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        self.connect('activate', lambda *ignore: Widgets.main.present())
        self.connect('command-line', command_line)
        self.connect('startup', startup)

        self.do_fade_in = do_fade_in

    def open_files(self, files):
        """Attempt to load all of the specified files.

        >>> len(Photograph.instances)
        0
        >>> GottenGeography().open_files(
        ...     ['demo/IMG_2411.JPG', 'demo/IMG_2412.JPG'])
        >>> len(Photograph.instances)
        2
        """
        Widgets.progressbar.show()
        invalid, total = [], len(files)
        for i, name in enumerate(files, 1):
            Widgets.redraw_interface(i / total, basename(name))
            try:
                try:
                    Photograph.load_from_file(name)
                except OSError:
                    TrackFile.load_from_file(name)
            except OSError:
                invalid.append(basename(name))
        if invalid:
            Widgets.status_message(_('Could not open: ') + ', '.join(invalid))

        # Ensure camera has found correct timezone regardless of the order
        # that the GPX/KML files were loaded in.
        likely_zone = TrackFile.query_all_timezones()
        if likely_zone:
            Camera.set_all_found_timezone(likely_zone)
        Camera.timezone_handler_all()
        Widgets.progressbar.hide()
        Widgets.button_sensitivity()

    def apply_selected_photos(self, button):
        """Manually apply map center coordinates to selected photos."""
        lat = MapView.get_center_latitude()
        lon = MapView.get_center_longitude()
        for photo in selected:
            photo.manual = True
            photo.set_location(lat, lon)
        Widgets.button_sensitivity()

    def save_all_files(self, *ignore):
        """Ensure all loaded files are saved."""
        Widgets.progressbar.show()
        total = len(modified)
        for i, photo in enumerate(list(modified), 1):
            Widgets.redraw_interface(i / total, basename(photo.filename))
            try:
                photo.write()
            except Exception as inst:
                Widgets.status_message(str(inst))
        Widgets.progressbar.hide()
        Widgets.button_sensitivity()

    def jump_to_photo(self, button):
        """Center on the first selected photo."""
        photo = selected.copy().pop()
        if photo.positioned:
            MapView.emit('realize')
            MapView.center_on(photo.latitude, photo.longitude)

    def open_help(self, *ignore):
        screen = Gdk.Screen.get_default()
        try:
            Gtk.show_uri(screen, 'help:gottengeography', Gdk.CURRENT_TIME),
        except GLib.GError:
            pass

    def map_source_menu(self, *ignore):
        screen = Gdk.Screen.get_default()
        lat, lon = MapView.get_center_latitude(), MapView.get_center_longitude()
        try:
            Gtk.show_uri(screen, 'http://maps.google.com/maps?q={},{}'.format(
                lat, lon), Gdk.CURRENT_TIME),
        except GLib.GError:
            pass

    def update_preview(self, chooser, image):
        """Display photo thumbnail and geotag data in file chooser."""
        image.set_from_stock(Gtk.STOCK_FILE, Gtk.IconSize.DIALOG)
        try:
            image.set_from_pixbuf(fetch_thumbnail(
                chooser.get_preview_filename(), 300))
        except (OSError, TypeError):
            return

    def add_files_dialog(self, *ignore):
        """Display a file chooser, and attempt to load chosen files."""
        response = Widgets.open.run()
        Widgets.open.hide()
        Widgets.redraw_interface()
        if response == Gtk.ResponseType.OK:
            self.open_files(Widgets.open.get_filenames())

    def confirm_quit_dialog(self, *ignore):
        """Teardown method, inform user of unsaved files, if any."""
        if not modified:
            self.quit()
            return True
        Widgets.quit.format_secondary_markup(self.quit_message % len(modified))
        response = Widgets.quit.run()
        Widgets.quit.hide()
        Widgets.redraw_interface()
        if response == Gtk.ResponseType.ACCEPT:
            self.save_all_files()
        if response != Gtk.ResponseType.CANCEL:
            self.quit()
        return True



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Reverse geocoding and other mathematical calculations."""


from gi.repository import GLib, GObject
from time import strftime, localtime
from gettext import gettext as _
from os.path import join

from gg.territories import get_state, get_country
from gg.build_info import PKG_DATA_DIR
from gg.common import memoize


def valid_coords(lat, lon):
    """Determine the validity of coordinates.

    >>> valid_coords(200, 300)
    False
    >>> valid_coords(40.689167, -74.044678)
    True
    >>> valid_coords(50, [])
    False
    """
    if {type(lat), type(lon)} - {float, int}:
        return False
    return abs(lat) <= 90 and abs(lon) <= 180


@memoize
def do_cached_lookup(key):
    """Scan cities.txt for the nearest town.

    >>> do_cached_lookup(GeoCacheKey(43.646424, -79.333426))
    ('Toronto', '08', 'CA', 'America/Toronto\\n')
    >>> do_cached_lookup(GeoCacheKey(48.440257, -89.204443))
    ('Thunder Bay', '08', 'CA', 'America/Thunder_Bay\\n')
    """
    near, dist = None, float('inf')
    lat1, lon1 = key.lat, key.lon
    with open(join(PKG_DATA_DIR, 'cities.txt')) as cities:
        for city in cities:
            name, lat2, lon2, country, state, tz = city.split('\t')
            x = (float(lon2) - lon1)
            y = (float(lat2) - lat1)
            delta = x * x + y * y
            if delta < dist:
                dist = delta
                near = (name, state, country, tz)
    return near


class GeoCacheKey:
    """This class allows fuzzy geodata cache lookups."""

    def __init__(self, lat, lon):
        self.key = '{:.2f},{:.2f}'.format(lat, lon)
        self.lat = lat
        self.lon = lon

    def __str__(self):
        """Show the key being used.

        >>> print(GeoCacheKey(53.564, -113.564))
        53.56,-113.56
        """
        return self.key

    def __hash__(self):
        """Different instances can be used to fetch dictionary values.

        >>> cache = { GeoCacheKey(53.564, -113.564): 'example' }
        >>> cache.get(GeoCacheKey(53.559, -113.560))
        'example'
        >>> cache.get(GeoCacheKey(0, 0), 'Missing')
        'Missing'
        """
        return hash(self.key)

    def __eq__(self, other):
        """Different instances can compare equally.

        >>> GeoCacheKey(10.004, 10.004) == GeoCacheKey(9.996, 9.996)
        True
        >>> GeoCacheKey(10.004, 10.004) == GeoCacheKey(0, 0)
        False
        """
        return self.key == other.key


class Coordinates(GObject.GObject):
    """A generic object containing latitude and longitude coordinates.

    >>> import os, time
    >>> os.environ['TZ'] = 'America/Winnipeg'
    >>> time.tzset()
    >>> coord = Coordinates()

    >>> coord.date
    >>> coord.timestamp = 60 * 60 * 24 * 365 * 50
    >>> coord.date
    '2019-12-19 06:00:00 PM'

    >>> coord.height
    >>> coord.altitude = 600
    >>> coord.height
    '600.0m above sea level'
    >>> coord.altitude = -100
    >>> coord.height
    '100.0m below sea level'

    >>> coord.positioned
    False
    >>> coord.coords
    >>> coord.latitude = -51.688687
    >>> coord.longitude = -57.804152
    >>> coord.coords
    'S 51.68869, W 57.80415'
    >>> coord.positioned
    True

    >>> coord.lookup_geodata()
    'Atlantic/Stanley'
    >>> coord.geoname
    'Stanley, Falkland Islands'
    """
    modified_timeout = None
    timeout_seconds = 0
    geotimezone = ''
    names = (None, None, None)

    timestamp = GObject.property(type=int)
    altitude = GObject.property(type=float)
    latitude = GObject.property(type=float, minimum=-90.0, maximum=90.0)
    longitude = GObject.property(type=float, minimum=-180.0, maximum=180.0)

    @GObject.property(type=bool, default=False)
    def positioned(self):
        """Identify if this instance occupies a valid point on the map.

        Returns False at 0,0 because it's actually remarkably difficult to
        achieve that exact point in a natural way (not to mention it's in the
        middle of the Atlantic), which means the photo hasn't been placed yet.
        """
        return bool(self.latitude or self.longitude)

    @GObject.property(type=str)
    def geoname(self):
        """Report the city, state, and country in a pretty list."""
        return ', '.join([name for name in self.names if name])

    @GObject.property(type=str)
    def date(self):
        """Convert epoch seconds to a human-readable date."""
        if self.timestamp:
            return strftime('%Y-%m-%d %X', localtime(self.timestamp))

    @GObject.property(type=str)
    def coords(self):
        """Report a nicely formatted latitude and longitude pair."""
        if self.positioned:
            lat, lon = self.latitude, self.longitude
            return '{} {:.5f}, {} {:.5f}'.format(
                _('N') if lat >= 0 else _('S'), abs(lat),
                _('E') if lon >= 0 else _('W'), abs(lon)
            )

    @GObject.property(type=str)
    def height(self):
        """Convert elevation into a human readable format."""
        if self.altitude:
            return '{:.1f}{}'.format(
                abs(self.altitude),
                _('m above sea level')
                if self.altitude >= 0 else
                _('m below sea level'))

    def __init__(self, **props):
        self.filename = ''

        GObject.GObject.__init__(self, **props)

        for prop in ('latitude', 'longitude', 'altitude', 'timestamp'):
            self.connect('notify::' + prop, self.do_modified)

    def __str__(self):
        """Plaintext summary of metadata.

        >>> coord = Coordinates()
        >>> print(coord)
        <BLANKLINE>
        >>> coord.altitude = 456.7
        >>> coord.latitude = 10
        >>> coord.lookup_geodata()
        'Africa/Accra'
        >>> print(coord)
        Yendi, Northern, Ghana
        N 10.00000, E 0.00000
        456.7m above sea level
        """
        values = (self.geoname, self.date, self.coords, self.height)
        return '\n'.join([val for val in values if val])

    def lookup_geodata(self):
        """Check the cache for geonames, and notify of any changes.

        >>> coord = Coordinates()
        >>> coord.lookup_geodata()
        >>> coord.latitude = 47.56494
        >>> coord.longitude = -52.70931
        >>> coord.lookup_geodata()
        'America/St_Johns'
        >>> coord.geoname
        "St. John's, Newfoundland and Labrador, Canada"
        """
        if not self.positioned:
            return

        old_geoname = self.geoname
        city, state, code, tz = do_cached_lookup(
            GeoCacheKey(self.latitude, self.longitude))
        self.names = (city, get_state(code, state), get_country(code))
        self.geotimezone = tz.strip()
        if self.geoname != old_geoname:
            self.notify('geoname')

        return self.geotimezone

    def do_modified(self, *ignore):
        """Set timer to update the geoname after all modifications are done.

        >>> coord = Coordinates()
        >>> type(coord.modified_timeout)
        <class 'NoneType'>
        >>> coord.latitude = 10
        >>> type(coord.modified_timeout)
        <class 'int'>
        """
        self.notify('positioned')
        self.notify('coords')
        if not self.modified_timeout:
            self.modified_timeout = GLib.timeout_add_seconds(
                self.timeout_seconds, self.update_derived_properties)

    def update_derived_properties(self):
        """Do expensive geodata lookups after the timeout.

        >>> coord = Coordinates()
        >>> coord.latitude = 10
        >>> type(coord.modified_timeout)
        <class 'int'>
        >>> coord.update_derived_properties()
        False
        >>> type(coord.modified_timeout)
        <class 'NoneType'>
        >>> coord.geoname
        'Yendi, Northern, Ghana'
        """
        if self.modified_timeout:
            self.notify('positioned')
            GLib.source_remove(self.modified_timeout)
            self.lookup_geodata()
            self.modified_timeout = None
        return False



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Control the drag & drop behavior.

This class allows three types of drags to happen:

1. Drags from an external source onto the left-hand photo pane.

2. Drags from an external source onto the map.

3. Drags from the left-hand photo pane onto the map.

In cases 2 & 3 where photos are dragged onto the map (regardless of the drag
source), the photos will be tagged with the precise map coordinates that they
were dragged to on the map. Otherwise the photos are simply loaded without
any modifications being made to the location tags.

Note that the code controlling dragging ChamplainLabels around within the map
is defined in label.py
"""


from gi.repository import Gtk, Gdk
from urllib.parse import urlparse, unquote

from gg.widgets import Widgets, MapView
from gg.common import selected
from gg.photos import Photograph


class DragController(object):
    """Control the drag & drop behavior."""

    def __init__(self, open_files):
        # Drag source definitons
        photos_view = Widgets.photos_view
        photos_view.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK, [], Gdk.DragAction.COPY)
        photos_view.drag_source_add_text_targets()
        photos_view.connect('drag-data-get', self.photo_drag_start)

        # Drag destination defintions
        notebook = Widgets.photo_camera_gps
        notebook.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        notebook.drag_dest_add_text_targets()
        notebook.connect('drag-data-received', self.photo_drag_end, False)

        container = Widgets.map_container
        container.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        container.drag_dest_add_text_targets()
        container.connect('drag-data-received', self.photo_drag_end, True)

        self.external_drag = True
        self.selection = photos_view.get_selection()
        self.open_files = open_files

    def photo_drag_start(self, widget, drag_context, data, info, time):
        """Allow dragging more than one photo."""
        self.external_drag = False  # Don't reload files from disk
        data.set_text('\n'.join([photo.filename for photo in selected]), -1)

    def photo_drag_end(self, widget, drag_context, x, y,
                       data, info, time, on_map):
        """Respond to drops and position photos accordingly.

        This method allows photos to be dropped in from the photo
        pane or any other drag source, such as the file browser.
        """
        if not data.get_text():
            return

        lat, lon = MapView.y_to_latitude(y), MapView.x_to_longitude(x)

        files = [unquote(urlparse(s).path.strip()) for s in
                 data.get_text().split('\n') if s]

        if self.external_drag:
            self.open_files(files)
        self.external_drag = True

        if on_map:
            for filename in files:
                photo = Photograph.cache.get(filename)
                if photo is not None:
                    photo.manual = True
                    photo.set_location(lat, lon)

        self.selection.emit('changed')



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Control how the map is navigated."""


from gi.repository import Gdk

from gg.common import Gst
from gg.gpsmath import valid_coords
from gg.widgets import Widgets, MapView


def move_by_arrow_keys(accel_group, acceleratable, keyval, modifier):
    """Move the map view by 5% of its length in the given direction."""
    key = Gdk.keyval_name(keyval)
    factor = 0.45 if key in ('Up', 'Left') else 0.55
    if key in ('Up', 'Down'):
        lat = MapView.y_to_latitude(MapView.get_height() * factor)
        lon = MapView.get_center_longitude()
    else:
        lat = MapView.get_center_latitude()
        lon = MapView.x_to_longitude(MapView.get_width() * factor)
    if valid_coords(lat, lon):
        MapView.center_on(lat, lon)


def remember_location(view):
    """Add current location to history stack."""
    history = list(Gst.get('history'))
    location = tuple([
        view.get_property(x) for x in ('latitude', 'longitude', 'zoom-level')])
    if history[-1] != location:
        history.append(location)
    Gst.set_history(history[-30:])


def go_back(*ignore):
    """Return the map view to where the user last set it."""
    history = list(Gst.get('history'))
    lat, lon, zoom = history.pop()
    if valid_coords(lat, lon):
        MapView.set_zoom_level(zoom)
        MapView.center_on(lat, lon)
    if len(history) > 1:
        Gst.set_history(history)
    else:
        Gst.reset('history')


def zoom_button_sensitivity(view, signal, inn_sensitive, out_sensitive):
    """Ensure zoom buttons are only sensitive when they need to be."""
    zoom = view.get_zoom_level()
    out_sensitive(view.get_min_zoom_level() != zoom)
    inn_sensitive(view.get_max_zoom_level() != zoom)


for prop in ('latitude', 'longitude', 'zoom-level'):
    Gst.bind(prop, MapView, prop)


MapView.connect(
    'notify::zoom-level',
    zoom_button_sensitivity,
    Widgets.zoom_in_button.set_sensitive,
    Widgets.zoom_out_button.set_sensitive)
MapView.connect('realize', remember_location)



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Custom map sources are defined here:

>>> MAP_SOURCES['osm-mapnik']                               #doctest: +ELLIPSIS
<MapSourceChain object at 0x... (ChamplainMapSourceChain at 0x...)>

Custom map actors are defined here as well:

>>> MapView.notify('width')
>>> Box.get_width() == MapView.get_width()
True
"""


from gi.repository import Gtk, Champlain, Clutter
from time import sleep

from gg.widgets import Widgets, MapView
from gg.common import Gst, singleton, memoize

START = Clutter.BinAlignment.START
CENTER = Clutter.BinAlignment.CENTER
END = Clutter.BinAlignment.END

MAP_SOURCES = {}
MAP_DESCS = [
    ['osm-mapnik', 'OpenStreetMap Mapnik', 0, 18, 256,
     'Map data is CC-BY-SA 2.0 OpenStreetMap contributors',
     'http://creativecommons.org/licenses/by-sa/2.0/',
     'http://tile.openstreetmap.org/#Z#/#X#/#Y#.png'],

    ['osm-cyclemap', 'OpenStreetMap Cycle Map', 0, 17, 256,
     'Map data is CC-BY-SA 2.0 OpenStreetMap contributors',
     'http://creativecommons.org/licenses/by-sa/2.0/',
     'http://a.tile.opencyclemap.org/cycle/#Z#/#X#/#Y#.png'],

    ['osm-transport', 'OpenStreetMap Transport Map', 0, 18, 256,
     'Map data is CC-BY-SA 2.0 OpenStreetMap contributors',
     'http://creativecommons.org/licenses/by-sa/2.0/',
     'http://tile.xn--pnvkarte-m4a.de/tilegen/#Z#/#X#/#Y#.png'],

    ['mff-relief', 'Maps for Free Relief', 0, 11, 256,
     'Map data available under GNU Free Documentation license, v1.2 or later',
     'http://www.gnu.org/copyleft/fdl.html',
     'http://maps-for-free.com/layer/relief/z#Z#/row#Y#/#Z#_#X#-#Y#.jpg'],
]

for map_desc in MAP_DESCS:
    mapid, name, min_zoom, max_zoom, size, lic, lic_uri, tile_uri = map_desc

    c = Champlain.MapSourceChain()
    c.push(Champlain.MapSourceFactory.dup_default().create_error_source(size))

    c.push(Champlain.NetworkTileSource.new_full(
        mapid, name, lic, lic_uri, min_zoom, max_zoom,
        size, Champlain.MapProjection.MERCATOR,
        tile_uri, Champlain.ImageRenderer()))

    c.push(Champlain.FileCache.new_full(1e8, None, Champlain.ImageRenderer()))
    c.push(Champlain.MemoryCache.new_full(100, Champlain.ImageRenderer()))
    MAP_SOURCES[mapid] = c


@memoize
class RadioMenuItem(Gtk.RadioMenuItem):
    """Create the individual menu items for choosing map sources.

    >>> RadioMenuItem(MAP_SOURCES['osm-cyclemap']).get_label()
    'OpenStreetMap Cycle Map'
    >>> len(Widgets.map_source_menu.get_children()) == len(MAP_SOURCES)
    True
    """

    def __init__(self, source):
        super().__init__()
        if self.instances:
            self.set_property('group', list(self.instances).pop())
        self.set_label(source.get_name())
        self.connect('activate', self.menu_item_clicked, source.get_id())
        Widgets.map_source_menu.append(self)

    def menu_item_clicked(self, item, map_id):
        """Switch to the clicked map source."""
        if self.get_active():
            MapView.set_map_source(MAP_SOURCES[map_id])


@singleton
class Sources:
    """Set up the source menu and link to GSettings."""

    def __init__(self):
        last_source = Gst.get_string('map-source-id')
        Gst.bind_with_convert(
            'map-source-id', MapView, 'map-source',
            MAP_SOURCES.get, lambda x: x.get_id())

        for source_id in sorted(MAP_SOURCES):
            menu_item = RadioMenuItem(MAP_SOURCES[source_id])
            if last_source == source_id:
                menu_item.set_active(True)

        Widgets.map_source_menu.show_all()


@singleton
class Crosshair(Champlain.Point):
    """Display a target at map center for placing photos."""

    def __init__(self):
        super().__init__()
        self.set_size(4)
        self.set_color(Clutter.Color.new(0, 0, 0, 64))
        Gst.bind('show-map-center', self, 'visible')
        MapView.bin_layout_add(self, CENTER, CENTER)


@singleton
class Scale(Champlain.Scale):
    """Display a distance scale on the map."""

    def __init__(self):
        super().__init__()
        self.connect_view(MapView)
        Gst.bind('show-map-scale', self, 'visible')
        MapView.bin_layout_add(self, START, END)


@singleton
class CoordLabel(Clutter.Text):
    """Put the current map coordinates into the black coordinate bar."""

    def __init__(self):
        super().__init__()
        self.set_color(Clutter.Color.new(255, 255, 255, 255))


@singleton
class Box(Clutter.Box):
    """Draw the black coordinate display bar atop map."""

    def __init__(self):
        super().__init__()
        self.set_layout_manager(Clutter.BinLayout())
        self.set_color(Clutter.Color.new(0, 0, 0, 96))
        Gst.bind('show-map-coords', self, 'visible')
        MapView.bin_layout_add(self, START, START)
        self.get_layout_manager().add(CoordLabel, CENTER, CENTER)
        MapView.connect(
            'notify::width',
            lambda *ignore: self.set_size(MapView.get_width(), 30))


def animate_in(anim=True):
    """Fade in all the map actors."""
    for i in range(Gst.get_int('animation-steps') if anim else 1, 0, -1):
        for actor in (Crosshair, Box, Scale):
            actor.set_opacity(256 - i)
        Widgets.redraw_interface()
        sleep(0.01)



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Common classes and datatypes used throughout the app.

The `selected` and `modified` set()s contain Photograph() instances, and
are frequently used for iteration and membership testing throughout the app.

The `points` dict maps epoch seconds to ChamplainCoordinate() instances. This
is used to place photos on the map by looking up their timestamps.
"""


from gi.repository import GObject, Gio, GLib
from functools import wraps

from gg.version import PACKAGE


# These variables are used for sharing data between classes
selected = set()
modified = set()
points = {}


def singleton(cls):
    """Decorate a class with @singleton when There Can Be Only One.

    >>> @singleton
    ... class Highlander: pass
    >>> Highlander() is Highlander() is Highlander
    True
    >>> id(Highlander()) == id(Highlander)
    True
    """
    class single(cls):
        def __call__(self):
            return self
    return single()


def memoize(obj):
    """General-purpose cache for classes, methods, and functions.

    Functions are cached by their arguments:

    >>> @memoize
    ... def doubler(foo):
    ...     print('performing expensive calculation...')
    ...     return foo * 2
    >>> doubler(50)
    performing expensive calculation...
    100
    >>> doubler(50)
    100

    Methods are also cached by all their arguments, including `self`, which
    means that different instances will not share their cache. This is
    primarily used in the Gtk.Builder subclasses, where we want to cache slow
    widget lookups, but we don't want different instances stepping on each
    other's widgets:

    >>> class WidgetFactory:
    ...     @memoize
    ...     def get_by_name(self, name):
    ...         print('Making new widget named', name)
    ...         return '<<{}>>'.format(name)
    >>> one = WidgetFactory()
    >>> one.get_by_name('bob')
    Making new widget named bob
    '<<bob>>'
    >>> one.get_by_name('bob')
    '<<bob>>'
    >>> two = WidgetFactory()
    >>> two.get_by_name('bob')
    Making new widget named bob
    '<<bob>>'

    Finally, class instantiations are also cached based on the arguments passed
    to the constructor:

    >>> @memoize
    ... class Memorable:
    ...     def __init__(self, foo): pass
    >>> Memorable('alpha') is Memorable('alpha')
    True
    >>> Memorable('alpha') is Memorable('beta')
    False
    >>> len(Memorable.instances)
    2
    """
    cache = obj.cache = {}
    obj.instances = cache.values()

    @wraps(obj)
    def memoizer(*args, **kwargs):
        """Do cache lookups and populate the cache in the case of misses."""
        key = args[0] if len(args) is 1 else args
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


class staticmethod(object):
    """Make @staticmethods play nice with @memoize.

    >>> @memoize
    ... class HasStatic:
    ...     @staticmethod
    ...     def do_something():
    ...         print('Invoked with no arguments.')
    >>> HasStatic.do_something()
    Invoked with no arguments.

    Without this class, the above example would raise
    TypeError: 'staticmethod' object is not callable
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        """Call the static method with no instance."""
        return self.func(*args, **kwargs)


@memoize
class Binding(GObject.Binding):
    """Make it easier to bind properties between GObjects."""

    def __init__(self, source, sprop, target, tprop=None,
                 flags=GObject.BindingFlags.DEFAULT):
        super().__init__(
            source=source, source_property=sprop,
            target=target, target_property=tprop or sprop,
            flags=flags)


class GSettings(Gio.Settings):
    """Override GSettings to be more useful to me."""
    get = Gio.Settings.get_value

    def __init__(self, schema='ca.' + PACKAGE, path=None):
        if path is not None:
            path = '/ca/{}/{}s/{}/'.format(PACKAGE, schema, path)
            schema = 'ca.{}.{}'.format(PACKAGE, schema)

        super().__init__(schema, path)

        # These are used to avoid infinite looping.
        self._ignore_key_changed = False
        self._ignore_prop_changed = True

    def bind(self, key, widget, prop='', flags=Gio.SettingsBindFlags.DEFAULT):
        """Don't make me specify the default flags every time."""
        Gio.Settings.bind(self, key, widget, prop or key, flags)

    def bind_with_convert(self, key, widget, prop, key_to_prop, prop_to_key):
        """Recreate g_settings_bind_with_mapping from scratch.

        This method was shamelessly stolen from John Stowers'
        gnome-tweak-tool on May 14, 2012.
        """
        def key_changed(settings, key):
            """Update widget property."""
            if self._ignore_key_changed:
                return
            self._ignore_prop_changed = True
            widget.set_property(prop, key_to_prop(self[key]))
            self._ignore_prop_changed = False

        def prop_changed(widget, param):
            """Update GSettings key."""
            if self._ignore_prop_changed:
                return
            self._ignore_key_changed = True  # pragma: no cover
            self[key] = prop_to_key(widget.get_property(prop))
            self._ignore_key_changed = False

        self.connect('changed::' + key, key_changed)
        widget.connect('notify::' + prop, prop_changed)
        key_changed(self, key)  # init default state


@singleton
class Gst(GSettings):
    """This is the primary GSettings instance for main app settings only.

    It cannot be used to access the relocatable schema, for that you'll have
    to create a new GSettings() instance.
    """

    def set_history(self, value):
        """Convert the map history to an array of tuples."""
        self.set_value('history', GLib.Variant('a(ddi)', value))

    def set_window_size(self, value):
        """Convert the window size to a pair of ints."""
        self.set_value('window-size', GLib.Variant('(ii)', value))

    def set_color(self, color):
        """Convert the GdkColor to a three-int tuple."""
        variant = GLib.Variant('(iii)', (color.red, color.green, color.blue))
        self.set_value('track-color', variant)


class Struct:
    """This is a generic object which can be assigned arbitrary attributes.

    >>> foo = Struct({'one': 2})
    >>> foo.one
    2
    >>> foo.four = 4
    >>> foo.four
    4
    """

    def __init__(self, attributes={}):
        self.__dict__.update(attributes)



# Author: Robert Park <robru@gottengeography.ca>, (C) 2010
# Copyright: See COPYING file included with this distribution.

"""Define classes used for parsing GPX and KML XML files."""


from xml.parsers.expat import ParserCreate, ExpatError
from gi.repository import Champlain, Clutter, Gtk, Gdk
from dateutil.parser import parse as parse_date
from collections import defaultdict, deque
from re import compile as re_compile
from gettext import gettext as _
from os.path import basename
from calendar import timegm
from time import perf_counter

from gg.camera import Camera
from gg.gpsmath import Coordinates
from gg.common import staticmethod
from gg.widgets import Widgets, Builder, MapView
from gg.common import GSettings, Gst, Struct, memoize, points


BOTTOM = Gtk.PositionType.BOTTOM
RIGHT = Gtk.PositionType.RIGHT


def make_clutter_color(color):
    """Generate a Clutter.Color from the currently chosen color.

    >>> color = make_clutter_color(Gdk.Color(32767, 65535, 32767))
    >>> (color.red, color.green, color.blue)
    (127, 255, 127)
    """
    return Clutter.Color.new(
        *[x / 256 for x in [color.red, color.green, color.blue, 49152]])


def track_color_changed(selection, polys):
    """Update the color of any loaded GPX tracks."""
    color = selection.get_color()
    Gst.set_color(color)
    one = make_clutter_color(color)
    two = one.lighten().lighten()
    for i, polygon in enumerate(polys):
        polygon.set_stroke_color(two if i % 2 else one)


class Polygon(Champlain.PathLayer):
    """Extend a Champlain.PathLayer to automate appending points.

    >>> poly = Polygon()
    >>> coord = poly.append_point(49.899754, -97.137494, None)
    >>> (coord.lat, coord.lon, coord.ele)
    (49.899754, -97.137494, 0.0)
    >>> coord = poly.append_point(53.529201, -113.499324, 1000)
    >>> (coord.lat, coord.lon, coord.ele)
    (53.529201, -113.499324, 1000.0)
    """

    def __init__(self):
        Champlain.PathLayer.__init__(self)
        self.set_stroke_width(4)
        MapView.add_layer(self)

    def append_point(self, latitude, longitude, elevation):
        """Simplify appending a point onto a polygon."""
        coord = Champlain.Coordinate.new_full(latitude, longitude)
        coord.lat = latitude
        coord.lon = longitude
        try:
            coord.ele = float(elevation)
        except (ValueError, TypeError):
            coord.ele = 0.0
        self.add_node(coord)
        return coord


class XMLSimpleParser:
    """A simple wrapper for the Expat XML parser."""

    def __init__(self, filename, root, watch, call_start, call_end):
        self.state = defaultdict(str)
        self.call_start = call_start
        self.call_end = call_end
        self.watchlist = watch
        self.rootname = root
        self.tracking = None
        self.element = None

        self.parser = ParserCreate()
        self.parser.StartElementHandler = self.element_root

        try:
            with open(filename, 'rb') as xml:
                self.parser.ParseFile(xml)
        except ExpatError:
            raise OSError

    def element_root(self, name, attributes):
        """Called on the root XML element, we check if it's the one we want."""
        if name != self.rootname:
            raise OSError
        self.parser.StartElementHandler = self.element_start

    def element_start(self, name, attributes):
        """Only collect the attributes from XML elements that we care about."""
        if not self.tracking:
            if name not in self.watchlist:
                return
            if self.call_start(name, attributes):
                # Start tracking this element, accumulate everything under it.
                self.tracking = name
                self.parser.CharacterDataHandler = self.element_data
                self.parser.EndElementHandler = self.element_end

        if self.tracking is not None:
            self.element = name
            self.state.update(attributes)

    def element_data(self, data):
        """Accumulate all data for an element.

        Expat can call this handler multiple times with data chunks.
        """
        if not data.strip():
            return
        self.state[self.element] += data

    def element_end(self, name, state=None):
        """When the tag closes, pass its data to the end callback and reset."""
        if name != self.tracking:
            return

        self.call_end(name, self.state)
        self.tracking = None
        self.state.clear()
        self.parser.CharacterDataHandler = None
        self.parser.EndElementHandler = None


class TrackFile():
    """Parent class for all types of GPS track files.

    Subclasses must implement at least element_end.
    """
    range = []
    parse = XMLSimpleParser
    instances = set()

    @staticmethod
    def update_range():
        """Ensure that TrackFile.range contains the correct info."""
        while TrackFile.range:
            TrackFile.range.pop()
        if not TrackFile.instances:
            Widgets.empty_trackfile_list.show()
        else:
            Widgets.empty_trackfile_list.hide()
            TrackFile.range.extend([min(points), max(points)])

    @staticmethod
    def get_bounding_box():
        """Determine the smallest box that contains all loaded polygons."""
        bounds = Champlain.BoundingBox.new()
        for trackfile in TrackFile.instances:
            for polygon in trackfile.polygons:
                bounds.compose(polygon.get_bounding_box())
        return bounds

    @staticmethod
    def query_all_timezones():
        """Try to determine the most likely timezone the user is in.

        First we check all TrackFiles for the timezone at their starting point,
        and if they are all identical, we report it. If they do not match, then
        the user must travel a lot, and then we simply have no idea what
        timezone is likely to be the one that their camera is set to.
        """
        zones = set()
        for trackfile in TrackFile.instances:
            zones.add(trackfile.start.geotimezone)
        return None if len(zones) != 1 else zones.pop()

    @staticmethod
    def clear_all(*ignore):
        """Forget all GPX data, start over with a clean slate."""
        for trackfile in list(TrackFile.instances):
            trackfile.destroy()

        points.clear()

    @staticmethod
    def load_from_file(uri):
        """Determine the correct subclass to instantiate.

        Also time everything and report how long it took. Raises OSError if
        the file extension is unknown, or no track points were found.
        """
        start_time = perf_counter()

        try:
            gpx = globals()[uri[-3:].upper() + 'File'](uri)
        except KeyError:
            raise OSError

        Widgets.status_message(
            _('%d points loaded in %.2fs.') %
            (len(gpx.tracks), perf_counter() - start_time), True)

        if len(gpx.tracks) < 2:
            return

        TrackFile.instances.add(gpx)
        MapView.emit('realize')
        MapView.set_zoom_level(MapView.get_max_zoom_level())
        MapView.ensure_visible(TrackFile.get_bounding_box(), False)

        TrackFile.update_range()
        Camera.set_all_found_timezone(gpx.start.geotimezone)

    def __init__(self, filename, root, watch):
        self.watchlist = watch
        self.filename = filename
        self.progress = Widgets.progressbar
        self.polygons = set()
        self.widgets = Builder('trackfile')
        self.append = None
        self.tracks = {}
        self.clock = perf_counter()

        self.gst = GSettings('trackfile', basename(filename))
        if self.gst.get_string('start-timezone') is '':
            # Then this is the first time this file has been loaded
            # and we should honor the user-selected global default
            # track color instead of using the schema-defined default
            self.gst.set_value('track-color', Gst.get_value('track-color'))

        self.gst.bind_with_convert(
            'track-color',
            self.widgets.colorpicker,
            'color',
            lambda x: Gdk.Color(*x),
            lambda x: (x.red, x.green, x.blue))

        self.widgets.trackfile_label.set_text(basename(filename))
        self.widgets.unload.connect('clicked', self.destroy)
        self.widgets.colorpicker.set_title(basename(filename))
        self.widgets.colorpicker.connect('color-set',
                                         track_color_changed,
                                         self.polygons)

        Widgets.trackfile_unloads_group.add_widget(self.widgets.unload)
        Widgets.trackfile_colors_group.add_widget(self.widgets.colorpicker)
        Widgets.trackfiles_group.add_widget(self.widgets.trackfile_label)

        self.parse(filename, root, watch, self.element_start, self.element_end)

        if not self.tracks:
            raise OSError('No points found')

        points.update(self.tracks)
        keys = self.tracks.keys()
        self.alpha = min(keys)
        self.omega = max(keys)
        self.start = Coordinates(
            latitude=self.tracks[self.alpha].lat,
            longitude=self.tracks[self.alpha].lon)

        self.gst.set_string('start-timezone', self.start.lookup_geodata())

        Widgets.trackfiles_view.add(self.widgets.trackfile_settings)

    def element_start(self, name, attributes=None):
        """Determine when new tracks start and create a new Polygon."""
        if name == self.watchlist[0]:
            polygon = Polygon()
            self.polygons.add(polygon)
            self.append = polygon.append_point
            self.widgets.colorpicker.emit('color-set')
            return False
        return True

    def element_end(self, name=None, state=None):
        """Occasionally redraw the screen so the user can see activity."""
        if perf_counter() - self.clock > .2:
            self.progress.pulse()
            while Gtk.events_pending():
                Gtk.main_iteration()
            self.clock = perf_counter()

    def destroy(self, button=None):
        """Die a horrible death."""
        for polygon in self.polygons:
            MapView.remove_layer(polygon)
        self.polygons.clear()
        self.widgets.trackfile_settings.destroy()
        del self.cache[self.filename]
        TrackFile.instances.discard(self)
        points.clear()
        for trackfile in TrackFile.instances:
            points.update(trackfile.tracks)
        TrackFile.update_range()


# GPX files use ISO 8601 dates, which look like 2010-10-16T20:09:13Z.
# This regex splits that up into a list like 2010, 10, 16, 20, 09, 13.
split = re_compile(r'[:T.Z-]').split


@memoize
class GPXFile(TrackFile):
    """Support for the open GPS eXchange format."""

    def __init__(self, filename):
        TrackFile.__init__(self, filename, 'gpx', ('trkseg', 'trkpt'))

    def element_end(self, name, state):
        """Collect and use all the parsed data."""
        try:
            timestamp = timegm(list(map(int, split(state['time'])[0:6])))
            lat = float(state['lat'])
            lon = float(state['lon'])
        except Exception as error:
            print(error)
            return

        self.tracks[timestamp] = self.append(lat, lon, state.get('ele'))

        TrackFile.element_end(self)


@memoize
class TCXFile(TrackFile):
    """Support for Garmin's Training Center XML."""

    def __init__(self, filename):
        TrackFile.__init__(self, filename,
                           'TrainingCenterDatabase', ('Track', 'Trackpoint'))

    def element_end(self, name, state):
        """Collect and use all the parsed data."""
        try:
            timestamp = timegm(list(map(int, split(state['Time'])[0:6])))
            lat = float(state['LatitudeDegrees'])
            lon = float(state['LongitudeDegrees'])
        except Exception as error:
            print(error)
            return

        self.tracks[timestamp] = self.append(
            lat, lon, state.get('AltitudeMeters'))

        TrackFile.element_end(self)


@memoize
class KMLFile(TrackFile):
    """Support for Google's Keyhole Markup Language.

    The KML parser is a little bit different than the other XML parsers,
    it uses a two-pass approach of loading points first, then marking them on
    the map in the second pass. This is because KML has an absurd requirement
    of allowing `when` tags to be decoupled from their associated `gx:coord`
    tags which makes the single-pass approach significantly slower.
    """

    def __init__(self, filename):
        self.whens = deque()
        self.coords = deque()

        TrackFile.__init__(self, filename,
                           'kml', ('gx:Track', 'when', 'gx:coord'))

    def element_start(self, name, attributes=None):
        """Make note of where new polygons start."""
        if name == self.watchlist[0]:
            self.whens.append(None)
            self.coords.append(None)
            return False
        return True

    def element_end(self, name, state):
        """Watch for complete pairs of when and gx:coord tags.

        This is accomplished by maintaining parallel deques of each tag.
        """
        if name == 'when':
            try:
                timestamp = timegm(parse_date(state['when']).utctimetuple())
            except Exception as error:
                print(error)
                return
            self.whens.append(timestamp)
        if name == 'gx:coord':
            self.coords.append(state['gx:coord'].split())
        TrackFile.element_end(self)

    def parse(self, filename, root, watch, start, end):
        """Trigger the first pass and then do the second pass."""
        # First pass
        TrackFile.parse(filename, root, watch, start, end)

        # Second pass
        TrackFile.element_start(self, 'gx:Track')

        whens = self.whens
        coords = self.coords
        tracks = self.tracks
        append = self.append

        while whens and coords:
            when = whens.popleft()
            coord = coords.popleft()
            try:
                tracks[when] = append(
                    float(coord[1]), float(coord[0]), coord[2])
            except TypeError:
                TrackFile.element_start(self, 'gx:Track')
                append = self.append
            else:
                TrackFile.element_end(self)


@memoize
class CSVFile(TrackFile):
    """Support for Google's MyTracks' Comma Separated Values format.

    This implementation ignores everything before the first line that contains
    the necessary column headers, allowing you to have any arbitrary preamble
    you like. Extra columns are harmlessly ignored. All "values" must be
    "quoted" with "double quotes".
    """
    columns = None

    def __init__(self, filename):
        TrackFile.__init__(
            self, filename, None,
            ('Segment', 'Latitude (deg)', 'Longitude (deg)', 'Time'))

    def parse(self, filename, root, watch, start, end):
        """Call the appropriate handler for each line of the file."""
        with open(filename) as lines:
            parse_line = re_compile(r'"([^"]*)",?').findall
            for line in lines:
                self.parse_header(parse_line(line), self.columns)

    def parse_header(self, state, columns, alt='Altitude (m)'):
        """Ignore as many lines as necessary until column headers are found."""
        try:
            self.columns = Struct({
                col.split(' ')[0].lower(): state.index(col)
                for col in self.watchlist})
        except ValueError:
            return

        self.columns.alt = state.index(alt) if alt in state else -1

        self.parse_header = self.parse_row

    def parse_row(self, state, col):
        """All subsequent lines contain one track point each."""
        try:
            if int(state[col.segment]) > len(self.polygons):
                self.element_start('Segment')

            timestamp = timegm(list(map(int, split(state[col.time])[0:6])))
            lat = float(state[col.latitude])
            lon = float(state[col.longitude])
        except Exception as error:
            print(error)
            return

        self.tracks[timestamp] = self.append(
            lat, lon, state[col.alt] if col.alt >= 0 else 0.0)

        TrackFile.element_end(self)



# Author: Robert Park <robru@gottengeography.ca>, (C) 2012
# Copyright: See COPYING file included with this distribution.

"""The Camera class handles per-camera configuration.

It uniquely identifies each camera model that the user owns and stores settings
such as what timezone to use and how wrong the camera's clock is. A
'relocatable' GSettings schema is used to persist this data across application
launches.

Note that the Cameras tab will only display the cameras responsible for
creating the currently loaded photos. This means that it's entirely possible
for GSettings to store a camera definition that isn't displayed in the UI.
Rest assured that your camera's settings are simply gone but not forgotten,
and if you want to see the camera in the camera list, you should load a photo
taken by that camera.
"""


from gi.repository import GObject, Gtk
from math import modf as split_float
from gettext import gettext as _
from time import tzset
from os import environ

from gg.common import staticmethod
from gg.widgets import Builder, Widgets
from gg.common import GSettings, Binding, memoize
from gg.territories import tz_regions, get_timezone


@memoize
class Camera(GObject.GObject):
    """Store per-camera configuration in GSettings.

    >>> from mock import Mock as Photo
    >>> cam = Camera('unknown_camera')
    >>> cam.add_photo(Photo())
    >>> cam.num_photos
    1
    >>> photo = Photo()
    >>> cam.add_photo(photo)
    >>> cam.num_photos
    2
    >>> cam.remove_photo(photo)
    >>> cam.num_photos
    1

    >>> Camera.generate_id({'Make': 'Nikon',
    ...                     'Model': 'Wonder Cam',
    ...                     'Serial': '12345'})
    ('12345_nikon_wonder_cam', 'Nikon Wonder Cam')
    >>> Camera.generate_id({})
    ('unknown_camera', 'Unknown Camera')

    >>> cam = Camera('canon_canon_powershot_a590_is')
    >>> cam.timezone_method = 'lookup'
    >>> environ['TZ']
    'America/Edmonton'
    >>> cam.timezone_method = 'offset'
    >>> environ['TZ'].startswith('UTC')
    True
    """
    offset = GObject.property(type=int, minimum=-3600, maximum=3600)
    utc_offset = GObject.property(type=str)
    found_timezone = GObject.property(type=str)
    timezone_method = GObject.property(type=str)
    timezone_region = GObject.property(type=str)
    timezone_city = GObject.property(type=str)

    @GObject.property(type=int)
    def num_photos(self):
        """Read-only count of the loaded photos taken by this camera."""
        return len(self.photos)

    @staticmethod
    def generate_id(info):
        """Identifies a camera by serial number, make, and model."""
        maker = info.get('Make', '').capitalize()
        model = info.get('Model', '')

        # Some makers put their name twice
        model = model if model.startswith(maker) else maker + ' ' + model

        camera_id = '_'.join(sorted(info.values())).lower().replace(' ', '_')

        return (camera_id.strip(' _') or 'unknown_camera',
                model.strip() or _('Unknown Camera'))

    @staticmethod
    def set_all_found_timezone(timezone):
        """"Set all cameras to the given timezone."""
        for camera in Camera.instances:
            camera.found_timezone = timezone

    @staticmethod
    def timezone_handler_all():
        """Update all of the photos from all of the cameras."""
        for camera in Camera.instances:
            camera.timezone_handler()

    def __init__(self, camera_id):
        GObject.GObject.__init__(self)
        self.id = camera_id
        self.photos = set()

        # Bind properties to settings
        self.gst = GSettings('camera', camera_id)
        for prop in self.gst.list_keys():
            self.gst.bind(prop, self)

        # Get notifications when properties are changed
        self.connect('notify::offset', self.offset_handler)
        self.connect('notify::timezone-method', self.timezone_handler)
        self.connect('notify::timezone-city', self.timezone_handler)
        self.connect('notify::utc-offset', self.timezone_handler)

    def timezone_handler(self, *ignore):
        """Set the timezone to the chosen zone and update all photos."""
        environ['TZ'] = ''
        if self.timezone_method == 'lookup':
            # Note that this will gracefully fallback on system timezone
            # if no timezone has actually been found yet.
            environ['TZ'] = self.found_timezone
        elif self.timezone_method == 'offset':
            minutes, hours = split_float(-float(self.utc_offset))
            environ['TZ'] = 'UTC{:+}:{:02}'.format(
                int(hours), int(abs(minutes) * 60))
        elif self.timezone_method == 'custom' and \
                self.timezone_region and self.timezone_city:
            environ['TZ'] = '/'.join(
                [self.timezone_region, self.timezone_city])

        tzset()
        self.offset_handler()

    def get_offset_from_clock_photo(self, btn, orig, tz):
        """Subtract the camera's clock from the photographed clock."""
        delta_s = Widgets.clock_photo_seconds.get_value_as_int() - orig.tm_sec
        delta_m = Widgets.clock_photo_minutes.get_value_as_int() - orig.tm_min
        delta_h = Widgets.clock_photo_hours.get_value_as_int() - orig.tm_hour
        delta_h += float(Widgets.clock_photo_tz.get_active_id())
        self.offset = delta_m * 60 + delta_s

        utc_offset = int(tz[1:3]) + int(tz[-2:]) / 60
        utc_offset *= -1 if tz.startswith('-') else 1
        self.utc_offset = str(utc_offset + delta_h)
        self.timezone_method = 'offset'

    def offset_handler(self, *ignore):
        """When the offset is changed, update the loaded photos."""
        for i, photo in enumerate(self.photos):
            if not i % 10:
                Widgets.redraw_interface()
            photo.calculate_timestamp(self.offset)

    def add_photo(self, photo):
        """Adds photo to the list of photos taken by this camera."""
        photo.camera = self
        self.photos.add(photo)
        self.notify('num_photos')

    def remove_photo(self, photo):
        """Removes photo from the list of photos taken by this camera."""
        photo.camera = None
        self.photos.discard(photo)
        self.notify('num_photos')


def display_offset(offset, value, add, subtract):
    """Display minutes and seconds in the offset GtkScale.

    >>> display_offset(None, 10, 'Add %d, %d', 'Subtract %d, %d')
    'Add 0, 10'
    >>> display_offset(None, 100, 'Add %d, %d', 'Subtract %d, %d')
    'Add 1, 40'
    >>> display_offset(None, -100, 'Add %d, %d', 'Subtract %d, %d')
    'Subtract 1, 40'
    """
    seconds, minutes = split_float(abs(value) / 60)
    return (subtract if value < 0 else add) % (minutes, int(seconds * 60))


@memoize
class CameraView(Gtk.Box):
    """A widget to show camera settings.

    >>> view = CameraView(Camera('unknown_camera'), 'Unknown Camera')
    >>> view.widgets.camera_label.get_text()
    'Unknown Camera'
    >>> view.widgets.timezone_method.set_active_id('system')
    True
    >>> view.widgets.utc_label.get_visible()
    False
    >>> view.camera.gst.get_string('timezone-method')
    'system'
    >>> view.widgets.timezone_method.set_active_id('offset')
    True
    >>> view.widgets.utc_label.get_visible()
    True
    >>> view.camera.gst.get_string('timezone-method')
    'offset'
    """

    def __init__(self, camera, name):
        Gtk.Box.__init__(self)
        self.camera = camera

        self.widgets = Builder('camera')
        self.add(self.widgets.camera_settings)

        self.widgets.camera_label.set_text(name)

        self.set_counter_text()

        # GtkScale allows the user to correct the camera's clock.
        scale = self.widgets.offset
        scale.connect('format-value', display_offset,
                      _('Add %dm, %ds to clock.'),
                      _('Subtract %dm, %ds from clock.'))

        # NOTE: This has to be so verbose because of
        # https://bugzilla.gnome.org/show_bug.cgi?id=675582
        # Also, it seems SYNC_CREATE doesn't really work.
        scale.set_value(camera.offset)
        Binding(scale.get_adjustment(), 'value', camera, 'offset',
                flags=GObject.BindingFlags.BIDIRECTIONAL)

        utc = self.widgets.utc_offset
        utc.set_active_id(camera.utc_offset)
        Binding(utc, 'active-id', camera, 'utc-offset',
                flags=GObject.BindingFlags.BIDIRECTIONAL)

        Binding(scale.get_adjustment(), 'value', camera, 'offset')

        # These two ComboBoxTexts are used for choosing the timezone manually.
        # They're hidden to reduce clutter when not needed.
        region_combo = self.widgets.timezone_region
        cities_combo = self.widgets.timezone_city
        for name in tz_regions:
            region_combo.append(name, name)

        for setting in ('region', 'city', 'method'):
            name = 'timezone_' + setting
            Binding(self.widgets[name], 'active-id',
                    camera, name.replace('_', '-'))

        region_combo.connect('changed', self.region_handler, cities_combo)
        region_combo.set_active_id(camera.timezone_region)
        cities_combo.set_active_id(camera.timezone_city)

        method_combo = self.widgets.timezone_method
        method_combo.connect('changed', self.method_handler)
        method_combo.set_active_id(camera.timezone_method)

        Widgets.timezone_cities_group.add_widget(utc)
        Widgets.timezone_cities_group.add_widget(cities_combo)
        Widgets.timezone_regions_group.add_widget(region_combo)
        Widgets.timezone_regions_group.add_widget(self.widgets.utc_label)
        Widgets.cameras_group.add_widget(self.widgets.camera_settings)

        camera.connect('notify::num-photos', self.set_counter_text)

        Widgets.cameras_view.add(self)
        self.show_all()

    def method_handler(self, method):
        """Only show manual tz selectors when necessary."""
        active_method = method.get_active_id()
        self.widgets.timezone_region.set_visible(active_method == 'custom')
        self.widgets.timezone_city.set_visible(active_method == 'custom')
        self.widgets.utc_label.set_visible(active_method == 'offset')
        self.widgets.utc_offset.set_visible(active_method == 'offset')

    def region_handler(self, region, cities):
        """Populate the list of cities when a continent is selected."""
        cities.remove_all()
        append = cities.append
        for city in get_timezone(region.get_active_id(), []):
            append(city.replace('_', ' '), city)

    def set_counter_text(self, *ignore):
        """Display to the user how many photos are loaded."""
        num = self.camera.num_photos
        self.widgets.count_label.set_text({
            0: _('No photos loaded.'),
            1: _('One photo loaded.'),
        }.get(num, _('%d photos loaded.') % num))
